# -*- coding: utf-8 -*-
##############################################################################
# Copyright (c) 2018 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
##############################################################################

import mock
import os
import time

from collections import namedtuple
from slapos.grid.promise import PromiseError
from slapos.promise.plugin.check_ram_usage import RunPromise
from . import TestPromisePluginMixin


class TestCheckRamUsage(TestPromisePluginMixin):

  promise_name = "monitor-ram-usage.py"

  def setUp(self):
    super(TestCheckRamUsage, self).setUp()
    self.ram_data = namedtuple('ram_data', ['available'])

  def writePromise(self, **kw):
    super(TestCheckRamUsage, self).writePromise(self.promise_name,
      "from %s import %s\nextra_config_dict = %r\n"
      % (RunPromise.__module__, RunPromise.__name__, kw))

  def runPromise(self, summary, failed=False):
    self.configureLauncher(enable_anomaly=True, force=True)
    with mock.patch('psutil.virtual_memory', return_value=summary):
      if failed:
        self.assertRaises(PromiseError, self.launcher.run)
      else:
        self.launcher.run()
    result = self.getPromiseResult(self.promise_name)['result']
    self.assertEqual(result['failed'], failed)
    return result['message']

  def test_ram_ok(self):
    message = "RAM usage OK"
    available_ram = {'available':1e9}
    self.writePromise(**{
        'last-avg-ram-file':'last_avg_ram_file',
        'min-threshold-ram': 500, # 500MB
        'min-avg-ram': 100,
    })
    self.assertEqual(message, self.runPromise(self.ram_data(**available_ram)))

  def test_ram_below_threshold_nok(self):
    message = "RAM usage reached critical threshold:  190.7M  (threshold is  500.0M)"
    available_ram = {'available': 200e6}
    self.writePromise(**{
        'last-avg-ram-file':'last_avg_ram_file',
        'min-threshold-ram': 500, # ≈500MB
        'min-avg-ram': 100,
    })
    self.assertEqual(message, self.runPromise(self.ram_data(**available_ram)))

  def test_ram_below_average_nok(self):
    message = "Average RAM usage over the last 1 seconds reached threshold:  190.7M (threshold is  200.0M)"
    available_ram = {'available': 200e6}
    self.writePromise(**{
        'last-avg-ram-file':'last_avg_ram_file',
        'min-threshold-ram': 0,
        'min-avg-ram': 200,
        'avg-ram-period': 1,
    })
    m = self.runPromise(self.ram_data(**{'available': 300e6}))
    m = self.runPromise(self.ram_data(**{'available': 200e6}))
    time.sleep(1)
    self.assertEqual(message, self.runPromise(self.ram_data(**available_ram)))
    

if __name__ == '__main__':
  unittest.main()



import json
import os
import psutil
import time

from psutil._common import bytes2human
from .util import JSONPromise

from zope.interface import implementer
from slapos.grid.promise import interface

@implementer(interface.IPromise)
class RunPromise(JSONPromise):

  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    # Get reference values
    self.setPeriodicity(float(self.getConfig('frequency', 1)))
    self.last_transit_file = self.getConfig('last-transit-file', 'last_transit')
    self.max_data_amount = float(self.getConfig('max-data-amount', 10e3))*1048576 #  MB converted into bytes
    self.min_data_amount = float(self.getConfig('min-data-amount', 0.1))*1048576 #  MB converted into bytes
    self.transit_period = int(self.getConfig('transit-period', 600)) # secondes

  def sense(self):
    promise_success = True    
    # Get current network statistics, see https://psutil.readthedocs.io/en/latest/#network
    network_data = psutil.net_io_counters(nowrap=True)
    data_amount = network_data.bytes_recv + network_data.bytes_sent
    # Log data amount
    data = json.dumps({'network_data_amount': data_amount})
    self.json_logger.info("Network data amount", extra={'data': data})

    # Get last timestamp (i.e. last modification) of log file
    try:
      t = os.path.getmtime(self.last_transit_file)
    except OSError:
      t = 0
    # We recalculate every quarter of transit_period since calculate over periodicity
    # can be heavy in computation
    if (time.time() - t) > self.transit_period / 4:
      open(self.last_transit_file, 'w').close()
      temp_list = self.get_json_log_data_interval(self.transit_period)
      if temp_list:
        # If no previous data in log
        if len(temp_list) == 1:
          pass
        else:
          data_diff = temp_list[0]['network_data_amount'] - temp_list[-1]['network_data_amount']
          if data_diff <= self.min_data_amount:
            self.logger.error("Network congested, data amount over the last %s seconds "\
              "reached minimum threshold: %7s (threshold is %7s)" 
              % (self.transit_period, bytes2human(data_diff), bytes2human(self.min_data_amount)))
            promise_success = False
          if data_diff >= self.max_data_amount:
            self.logger.error("Network congested, data amount over the last %s seconds "\
              "reached maximum threshold: %7s (threshold is %7s)" 
              % (self.transit_period, bytes2human(data_diff), bytes2human(self.max_data_amount)))
            promise_success = False
      else:
        self.logger.error("Couldn't read network data from log")
        promise_success = False

    if promise_success:
      self.logger.info("Network transit OK")

  def test(self):
    """
    Called after sense() if the instance is still converging.
    Returns success or failure based on sense results.

    In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)


  def anomaly(self):
    """
    Called after sense() if the instance has finished converging.
    Returns success or failure based on sense results.
    Failure signals the instance has diverged.

    In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=3, failure_amount=2)



from __future__ import division

from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise

import os
import sys
import pwd

import sqlite3
import argparse
import datetime
import psutil
import math
import pkgutil

# try to install pandas and numpy
try:
  import pandas as pd
  import numpy as np
except ImportError:
  pass

from slapos.collect.db import Database
from contextlib import closing

@implementer(interface.IPromise)
class RunPromise(GenericPromise):

  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    # at least every hours (heavy in computation)
    self.setPeriodicity(float(self.getConfig('frequency', 60)))

  def getDiskSize(self, disk_partition, db_path):
    database = Database(db_path, create=False, timeout=10)
    # by using contextlib.closing, we don't need to close the database explicitly
    with closing(database):
      try:
        database.connect()
        where_query = "partition='%s'" % (disk_partition)
        order = "datetime(date || ' ' || time) DESC"
        result = database.select(
          "disk",
          columns="free+used",
          where=where_query,
          order=order,
          limit=1).fetchone()
        if not result or not result[0]:
          return None
        disk_size = result[0]
      except sqlite3.OperationalError as e:
        # if database is still locked after timeout expiration (another process is using it)
        # we print warning message and try the promise at next run until max warn count
        locked_message = "database is locked"
        if locked_message in str(e) and \
            not self.raiseOnDatabaseLocked(locked_message):
          return None
        raise
    return disk_size

  def getPartitionSize(self, disk_partition, db_path):
    database = Database(db_path, create=False, timeout=10)
    with closing(database):
      try:
        database.connect()
        where_query = "partition='%s'" % (disk_partition)
        order = "datetime(date || ' ' || time) DESC"
        result = database.select(
          "folder",
          columns="disk_used*1024",
          where=where_query,
          order=order,
          limit=1).fetchone()
        if not result or not result[0]:
          return None
        partition_size = result[0]
      except sqlite3.OperationalError as e:
        # if database is still locked after timeout expiration (another process is using it)
        # we print warning message and try the promise at next run until max warn count
        locked_message = "database is locked"
        if locked_message in str(e) and \
            not self.raiseOnDatabaseLocked(locked_message):
          return None
        raise
    return partition_size

  def getAnomaly(self, disk_partition, db_path, user, date, time):
    database = Database(db_path, create=False, timeout=10)
    with closing(database):
      try:
        disk_size = self.getDiskSize(disk_partition, db_path)
        if disk_size is None:
          return None
        database.connect()
        result = database.select(
          "folder",
          columns = "%s-disk_used*1024, disk_used*1024, datetime(date || ' ' || time)" % disk_size,
          where =  "partition='%s'" % (user),
          order = "date ASC, time ASC"
        ).fetchall()
        if not result or not result[0]:
          self.logger.info("No result from collector database for the user %s: skipped", user)
          return None
        datetime_now = datetime.datetime.strptime(date + ' ' + time, "%Y-%m-%d %H:%M:%S")
        # check that the last data is less than 24 hours old
        last_date = datetime.datetime.strptime(result[-1][2], "%Y-%m-%d %H:%M:%S")
        if (datetime_now - last_date) > datetime.timedelta(days=1):
          self.logger.info("Not enough recent data to detect anomalies: skipped")
          return None
        # check that the first data is at least 13 days old
        first_date = datetime.datetime.strptime(result[0][2], "%Y-%m-%d %H:%M:%S")
        if (datetime_now - first_date) < datetime.timedelta(days=13):
          self.logger.info("Not enough data to detect anomalies: skipped")
          return None

        df = pd.DataFrame(result, columns=["free", "used", "date"])
        df.loc[:,'date'] = pd.to_datetime(df.date)
        # keep a sample every 5 minutes, set NaN when there is no information
        freq = 5
        df = df.resample(str(freq)+"min", on='date').mean()
        # estimate the missing information
        df['free'] = df.free.astype(float).interpolate(method='linear')
        df['used'] = df.used.astype(float).interpolate(method='linear')
        # calculate the median for the element-wise absolute value
        # of the difference between each x and the median of x
        df = df.reset_index()
        x = df['date']
        y = df['free']
        mad = lambda x: np.median(np.fabs(x - np.median(x)))
        # threshold is set at 8% of the disk size by default
        threshold_ratio = float(self.getConfig('threshold-ratio', 0.08) or 0.08)
        threshold = threshold_ratio*disk_size
        # use a 1-day window
        minutes_per_day = 60*24/freq
        rolling_window = int(minutes_per_day*1)
        rolling_mad = y.rolling(window=rolling_window, center=False).median() + \
          y.rolling(window=rolling_window, center=False).apply(mad)
        rolling_mad_upper = rolling_mad + threshold
        rolling_mad_lower = rolling_mad - threshold
        # create Pandas DataFrame and rename columns
        data = pd.concat([x, y, df['used'], rolling_mad, rolling_mad_upper, rolling_mad_lower], axis=1)
        data.columns = ["date", "free", "used", "mad", "upper_mad", "lower_mad"]
        # drop initial values (outside rolling window)
        data.dropna(subset=["mad"], inplace=True)
        # determine anomalies and display their number
        data["is_anomaly"] = ~(data["free"].between(data["lower_mad"], data["upper_mad"]))
        data = data.set_index("date")
        if (len(data)==0):
          self.logger.info("No result from anomaly detection")
          return None
        self.logger.info("There were %s anomalies in the last 15 days " \
          "(1 data every %s minutes, threshold: %s %% of the disk size)" % (
          len(data[data['is_anomaly'] == True]), freq, threshold_ratio*100))
        return data
      except sqlite3.OperationalError as e:
        # if database is still locked after timeout expiration (another process is using it)
        # we print warning message and try the promise at next run until max warn count
        locked_message = "database is locked"
        if locked_message in str(e) and \
            not self.raiseOnDatabaseLocked(locked_message):
          return None
        raise

  @staticmethod
  def _checkInodeUsage(path):
    stat = os.statvfs(path)
    total_inode = stat.f_files
    if total_inode:
      usage = 100 * (total_inode - stat.f_ffree) / total_inode
      if usage >= 98:
        return "Disk Inodes usage is really high: %.4f%%" % usage

  def getInodeUsage(self, path):
    return (self._checkInodeUsage(path) or
       os.path.ismount('/tmp') and self._checkInodeUsage('/tmp') or
       "")

  def sense(self):
    # check that the libraries are installed from the slapos.toolbox extra requires
    pandas_found = pkgutil.find_loader("pandas")
    numpy_found = pkgutil.find_loader("numpy")
    if pandas_found is None or numpy_found is None:
      self.logger.warning("Trying to use pandas but the module is not installed. Promise skipped.")
      return
    # find if a disk is mounted on the path
    disk_partition = ""
    db_path = self.getConfig('collectordb')
    check_date = self.getConfig('test-check-date')
    path = os.path.join(self.getPartitionFolder(), "") + "extrafolder"
    partitions = psutil.disk_partitions()
    while path is not '/':
      if not disk_partition:
        path = os.path.dirname(path)
      else:
        break
      for p in partitions:
        if p.mountpoint == path:
          disk_partition = p.device
          break
    if not disk_partition:
      self.logger.error("Couldn't find disk partition")
      return

    if db_path.endswith("collector.db"):
      db_path=db_path[:-len("collector.db")]

    if check_date:
      # testing mode
      currentdate = check_date
      currenttime = self.getConfig('test-check-time', '09:30:30')
      user = self.getConfig('test-partition', 'slapuser0')
    else:
      # get the user name of the partition
      user = pwd.getpwuid(os.getuid()).pw_name
      # get last minute
      now = datetime.datetime.utcnow()
      currentdate = now.strftime('%Y-%m-%d')
      currenttime = now - datetime.timedelta(minutes=1)
      currenttime = currenttime.time().strftime('%H:%M:%S')

    partition_size = self.getPartitionSize(user, db_path)
    data = self.getAnomaly(disk_partition, db_path, user, currentdate, currenttime)
    if data is None:
      return
    last_data = data.iloc[-1]
    last_date = data.index[-1]
    if last_data.is_anomaly:
      self.logger.error("Anomaly detected on %s. Space used by %s: %.2f G." % (
        last_date, user, partition_size/(1024*1024*1024)))
    else:
      self.logger.info("No anomaly detected (last date checked: %s)" % (last_date))

  def test(self):
    return self._test(result_count=1, failure_amount=1)

  def anomaly(self):
    return self._test(result_count=3, failure_amount=3)



##############################################################################
#
# Copyright (c) 2018 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from slapos.test.promise.plugin import TestPromisePluginMixin
from slapos.grid.promise import PromiseError
import os
import sqlite3
import psutil
from slapos.grid.promise import PromiseError

class TestMonitorPartitionSpace(TestPromisePluginMixin):

  def setUp(self):
    TestPromisePluginMixin.setUp(self)
    log_folder = os.path.join(self.partition_dir, 'var/log')
    os.makedirs(log_folder)

    # get disk partition name
    disk_partition = ""
    path = os.path.join(self.partition_dir, "") + "extrafolder"
    partitions = psutil.disk_partitions()
    while path is not '/':
      if not disk_partition:
        path = os.path.dirname(path)
      else:
        break
      for p in partitions:
        if p.mountpoint == path:
          disk_partition = p.device
          break

    self.db_file = '/tmp/collector.db'

    # populate db
    self.conn = sqlite3.connect(self.db_file)
    f = open(self.base_path+"/folder_disk_test.sql")
    sql = f.read()
    # replace every disk_partition_name string with disk partition name
    sql = sql.replace('disk_partition_name', disk_partition)
    self.conn.executescript(sql)
    self.conn.close()

    self.promise_name = "monitor-partition-space.py"

    content = """from slapos.promise.plugin.monitor_partition_space import RunPromise

extra_config_dict = {
  'collectordb': '%(collectordb)s',
  'test-check-date': '2017-10-02',
}
""" % {'collectordb': self.db_file}
    self.writePromise(self.promise_name, content)

  def tearDown(self):
    TestPromisePluginMixin.tearDown(self)
    if os.path.exists(self.db_file):
      os.remove(self.db_file)

  def test_no_data_for_a_partition(self):
    content = """from slapos.promise.plugin.monitor_partition_space import RunPromise

extra_config_dict = {
  'collectordb': '%(collectordb)s',
  'test-check-date': '2017-10-02',
  'test-partition': 'slapuser1'
}
""" % {'collectordb': self.db_file}

    self.writePromise(self.promise_name, content)

    self.configureLauncher(timeout=20)
    self.launcher.run()
    result = self.getPromiseResult(self.promise_name)
    self.assertEqual(result['result']['failed'], False)
    self.assertEqual(result['result']['message'], 
      "No result from collector database for the user slapuser1: skipped")

  def test_no_recent_data(self):
    content = """from slapos.promise.plugin.monitor_partition_space import RunPromise

extra_config_dict = {
  'collectordb': '%(collectordb)s',
  'test-check-date': '2017-10-04'
}
""" % {'collectordb': self.db_file}

    self.writePromise(self.promise_name, content)

    self.configureLauncher(force=True, timeout=20)
    self.launcher.run()
    result = self.getPromiseResult(self.promise_name)
    self.assertEqual(result['result']['failed'], False)
    self.assertEqual(result['result']['message'], "Not enough recent data to detect anomalies: skipped")

  def test_no_enough_data(self):
    content = """from slapos.promise.plugin.monitor_partition_space import RunPromise

extra_config_dict = {
  'collectordb': '%(collectordb)s',
  'test-check-date': '2017-09-24'
}
""" % {'collectordb': self.db_file}

    self.writePromise(self.promise_name, content)

    self.configureLauncher(force=True, timeout=20)
    self.launcher.run()
    result = self.getPromiseResult(self.promise_name)
    self.assertEqual(result['result']['failed'], False)
    self.assertEqual(result['result']['message'], 
      "Not enough data to detect anomalies: skipped")

  def test_no_anomalies(self):
    content = """from slapos.promise.plugin.monitor_partition_space import RunPromise

extra_config_dict = {
  'collectordb': '%(collectordb)s',
  'test-check-date': '2017-10-02',
  'threshold-ratio': '0.70'
}
""" % {'collectordb': self.db_file}

    self.writePromise(self.promise_name, content)

    self.configureLauncher(force=True, timeout=20)
    self.launcher.run()
    result = self.getPromiseResult(self.promise_name)
    self.assertEqual(result['result']['failed'], False)
    self.assertIn("No anomaly detected (last date checked: 2017-10-02 09:30:00)",
                  result['result']['message'])

  def test_presence_of_anomalies(self):
    self.configureLauncher(force=True, timeout=20)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    result = self.getPromiseResult(self.promise_name)
    self.assertEqual(result['result']['failed'], True)
    msg = "Anomaly detected on 2017-10-02 09:30:00. Space used by slapuser0: %.2f G."
    self.assertIn(msg % (87533020.0/(1024*1024)), result['result']['message'])


if __name__ == '__main__':
  unittest.main()


from .util import get_json_log_data_interval
from .util import JSONPromise

from zope.interface import implementer
from slapos.grid.promise import interface

@implementer(interface.IPromise)
class RunPromise(JSONPromise):

  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    self.setPeriodicity(minute=1)
    self.testing = self.getConfig('testing') == "True"
    self.amarisoft_stats_log = self.getConfig('amarisoft-stats-log')
    self.stats_period = int(self.getConfig('stats-period'))
    self.max_rx_sample_db = float(self.getConfig('max-rx-sample-db'))

  def sense(self):

    data_list = get_json_log_data_interval(self.amarisoft_stats_log, self.stats_period * 2)

    max_rx_list = []
    saturated = False

    for rx_antenna_list in map(lambda x: x['samples']['rx'], data_list):
        rx_list = map(lambda x: float(x['max']), rx_antenna_list)
        if not max_rx_list:
          max_rx_list = list(rx_list)
        for i, rx in enumerate(rx_list):
            max_rx_list[i] = max(max_rx_list[i], rx)
            if rx >= self.max_rx_sample_db:
                saturated = True

    self.json_logger.info("RX maximum sample values (dB)", extra={'data': max_rx_list})

    if not max_rx_list:
        self.logger.error("No RX samples data available")
    elif saturated:
        self.logger.error("RX antenna saturated, please lower rx_gain")
    else:
        self.logger.info("No saturation detected on RX antenna")

  def test(self):
    """
      Called after sense() if the instance is still converging.
      Returns success or failure based on sense results.

      In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)

  def anomaly(self):
    """
      Called after sense() if the instance has finished converging.
      Returns success or failure based on sense results.
      Failure signals the instance has diverged.

      In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=1, failure_amount=1)



import time

from .util import get_json_log_latest_timestamp

from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise

@implementer(interface.IPromise)
class RunPromise(GenericPromise):

  def __init__(self, config):

    super(RunPromise, self).__init__(config)
    self.setPeriodicity(minute=1)
    self.amarisoft_stats_log = self.getConfig('amarisoft-stats-log')
    self.stats_period = int(self.getConfig('stats-period'))

  def sense(self):

    latest_timestamp = get_json_log_latest_timestamp(self.amarisoft_stats_log)
    delta = time.time() - latest_timestamp
    if delta > self.stats_period * 2:
        self.logger.error("Latest entry from amarisoft statistics log too "\
                          "old (%s seconds old)" % (delta,))
    else:
        self.logger.info("Latest entry from amarisoft statistics is "\
                         "%s seconds old" % (delta,))

  def test(self):
    """
      Called after sense() if the instance is still converging.
      Returns success or failure based on sense results.

      In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)

  def anomaly(self):
    """
      Called after sense() if the instance has finished converging.
      Returns success or failure based on sense results.
      Failure signals the instance has diverged.

      In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=1, failure_amount=1)



import psutil
import math

from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise

@implementer(interface.IPromise)
class RunPromise(GenericPromise):

  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    # Get reference values
    self.setPeriodicity(float(self.getConfig('frequency', 5)))
    self.max_lost_packets = int(self.getConfig('max-lost-packets-per-MB', 100))
    self.max_error_messages = int(self.getConfig('max-error-messages-per-MB', 100))

  def sense(self):
    promise_success = True
    # Get current Network statistics
    network_data =  psutil.net_io_counters()
    # Get total number of bytes recv and sent in MB (if > 1MB)
    if (network_data.bytes_recv + network_data.bytes_sent) > 1e6:
      total_MB = (network_data.bytes_recv + network_data.bytes_sent)/1e6
    else:
      total_MB = 1
    # Get sum of errors and dropped packets
    total_dropped = network_data.dropin + network_data.dropout
    total_errors = network_data.errin + network_data.errout

    # Check for network dropped packets
    if total_dropped/total_MB >= self.max_lost_packets:
      self.logger.error("Network packets lost reached critical threshold: %s "\
        " (threshold is %s per MB)" % (math.ceil(total_dropped/total_MB), self.max_lost_packets))
      promise_success = False

    # Check for network errors
    if total_errors/total_MB >= self.max_error_messages:
      self.logger.error("Network errors reached critical threshold: %s "\
        " (threshold is %s per MB)" % (math.ceil(total_errors/total_MB), self.max_error_messages))
      promise_success = False

    if promise_success:
      self.logger.info("Network statistics OK")

  def test(self):
    """
    Called after sense() if the instance is still converging.
    Returns success or failure based on sense results.

    In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)


  def anomaly(self):
    """
    Called after sense() if the instance has finished converging.
    Returns success or failure based on sense results.
    Failure signals the instance has diverged.

    In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=3, failure_amount=2)
